/*
 * Server-side event management
 *
 * Copyright (C) 1998 Alexandre Julliard
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "winnt.h"

#include "handle.h"
#include "thread.h"
#include "request.h"

struct event
{
    struct object  obj;             /* object header */
    int            manual_reset;    /* is it a manual reset event? */
    int            write_fd;        /* fd to write to to set the event */
};

static void event_dump( struct object *obj, int verbose );
static int event_satisfied( struct object *obj, struct thread *thread );
static int event_get_poll_events( struct object *obj );
static int event_get_fd( struct object *obj );

static const struct object_ops event_ops =
{
    sizeof(struct event),      /* size */
    event_dump,                /* dump */
    default_poll_add_queue,    /* add_queue */
    default_poll_remove_queue, /* remove_queue */
    default_poll_signaled,     /* signaled */
    event_satisfied,           /* satisfied */
    event_get_poll_events,     /* get_poll_events */
    default_poll_event,        /* poll_event */
    event_get_fd,              /* get_fd */
    no_flush,                  /* flush */
    no_get_file_info,          /* get_file_info */
    no_destroy                 /* destroy */
};

static int event_get_poll_events( struct object *obj )
{
    assert (obj->ops == &event_ops);
    return POLLIN;
}

static int event_get_fd( struct object *obj )
{
    assert (obj->ops == &event_ops);
    return obj->fd;
}

struct event *create_event( const WCHAR *name, size_t len,
                            int manual_reset, int initial_state )
{
    struct event *event;

    if ((event = create_named_object( &event_ops, name, len )))
    {
        if (get_error() != STATUS_OBJECT_NAME_COLLISION)
        {
            /* initialize it if it didn't already exist */
            int fds[2];
            char data[1];

            fds[0]=0;
            fds[1]=0;
            if(0>pipe(&fds[0]))
                fprintf(stderr,"Couldn't open pipe for event flag.\n");
            if(0>fcntl(fds[0],F_SETFL,O_NONBLOCK))
                fprintf(stderr,"Couldn't make event flag pipe non-blocking.\n");
            event->obj.fd       = fds[0];
            event->write_fd     = fds[1];

            add_select_user(&event->obj);
            if(initial_state)
                write(fds[1], data, sizeof data);

            event->manual_reset = manual_reset;
        }
    }
    return event;
}

struct event *get_event_obj( struct process *process, handle_t handle, unsigned int access )
{
    return (struct event *)get_handle_obj( process, handle, access, &event_ops );
}

void pulse_event( struct event *event )
{
    set_event(event);
    wake_up( &event->obj, !event->manual_reset );
    reset_event(event);
}

void set_event( struct event *event )
{
    char data[1];

    data[0]=0;
    if(0>write(event->write_fd, &data, 1))
        fprintf(stderr,"event: write failed\n");
    wake_up(&event->obj, !event->manual_reset);
}

void reset_event( struct event *event )
{
    char data[32];

    /* make sure it doesn't block */
    while(1)
    {
        int r = read(event->obj.fd, &data[0], sizeof data);
        if(errno == EAGAIN)
             break;
        if(r<0)
        {
             perror("event");
             break;
        }
    }
}

static void event_dump( struct object *obj, int verbose )
{
    struct event *event = (struct event *)obj;
    assert( obj->ops == &event_ops );
    fprintf( stderr, "Event manual=%d write_fd=%d ",
             event->manual_reset, event->write_fd );
    dump_object_name( &event->obj );
    fputc( '\n', stderr );
}

static int event_satisfied( struct object *obj, struct thread *thread )
{
    struct event *event = (struct event *)obj;
    assert( obj->ops == &event_ops );
    /* Reset if it's an auto-reset event */
    if (!event->manual_reset) reset_event(event);
    return 0;  /* Not abandoned */
}

/* create an event */
DECL_HANDLER(create_event)
{
    struct event *event;

    req->handle = 0;
    if ((event = create_event( get_req_data(req), get_req_data_size(req),
                               req->manual_reset, req->initial_state )))
    {
        req->handle = alloc_handle( current->process, event, EVENT_ALL_ACCESS, req->inherit );
        release_object( event );
    }
}

/* open a handle to an event */
DECL_HANDLER(open_event)
{
    req->handle = open_object( get_req_data(req), get_req_data_size(req),
                               &event_ops, req->access, req->inherit );
}

/* do an event operation */
DECL_HANDLER(event_op)
{
    struct event *event;

    if (!(event = get_event_obj( current->process, req->handle, EVENT_MODIFY_STATE ))) return;
    switch(req->op)
    {
    case PULSE_EVENT:
        pulse_event( event );
        break;
    case SET_EVENT:
        set_event( event );
        break;
    case RESET_EVENT:
        reset_event( event );
        break;
    default:
        fatal_protocol_error( current, "event_op: invalid operation %d\n", req->op );
    }
    release_object( event );
}
